"""
gradients of excitation energies
"""
import numpy as np
import numpy.linalg as la

class Gradients:
    def __init__(self, tddftb):
        self.tddftb = tddftb
        self.dftb = tddftb.dftb2
        
    def getExcGradients(self, I=0):
        """
        compute gradients of excitation energies
        I=0 is the first excited state, not the ground state!

        according to "Analytical Excited State Forces for the TD-DFTB Method"
        by Heringer et al., J. Comput. Chem., (2007), 28, 16. 
        """
        XmY, XpY = self.tddftb.getXY()
        Omega = self.tddftb.getExcEnergies()

        err = abs(np.sum(XmY[I,:,:]*XpY[I,:,:]) - 1.0)
        assert err < 1.0e-10, "err = %s" % err
        S, H0 = self.dftb.getH0andS()
        orbs = self.dftb.getKSCoefficients()
        orbe = self.dftb.getKSEnergies()
        gamma = self.dftb.getGamma()
        dq = self.dftb.getPartialCharges()
        qoo,qvv,qov = self.tddftb.getTransitionCharges()

        occ, virt = self.tddftb.getActiveOrbitals()
        nocc, nvirt = len(occ), len(virt)
        norb = S.shape[0]
        nat,nat = gamma.shape
        at2orbs = self.dftb.getOrbitalsOnAtom()
        # orbs2at[p] gives the index of the atom on which orbital p sits
        orbs2at = np.zeros(norb,dtype=int)
        for A in range(0, nat):
            for m in at2orbs[A]:
                orbs2at[m] = A
        # LAGRANGE MULTIPLIERS Z and W
        ########################################
        # compute T, eqns. (B4)-(B8)
        # occ-occ block
        Too = np.zeros((nocc,nocc))
        for i in range(0, nocc):
            for j in range(0, nocc):
                Too[i,j] = np.sum(XpY[I,i,:]*XpY[I,j,:] + (XmY[I,i,:]*XmY[I,j,:]))
        Too *= -0.5
        # virt-virt block
        Tvv = np.zeros((nvirt,nvirt))
        for a in range(0, nvirt):
            for b in range(0, nvirt):
                Tvv[a,b] = np.sum(XpY[I,:,a]*XpY[I,:,b] + (XmY[I,:,a]*XmY[I,:,b]))
        Tvv *= 0.5

        zov = np.zeros((nocc,nvirt))
        zvo = np.zeros((nvirt,nocc))
        T = np.bmat([[Too,zov],[zvo,Tvv]])
        T = np.array(T)
        # compute coupling matrices necessary for Q
        Kvvov = np.tensordot(qvv, np.tensordot(gamma, qov, axes=(1,0)),axes=(0,0))
        Kooov = np.tensordot(qoo, np.tensordot(gamma, qov, axes=(1,0)),axes=(0,0))
        Kovvv = np.tensordot(qov, np.tensordot(gamma, qvv, axes=(1,0)),axes=(0,0))
        Kovoo = np.tensordot(qov, np.tensordot(gamma, qoo, axes=(1,0)),axes=(0,0))
        Koooo = np.tensordot(qoo, np.tensordot(gamma, qoo, axes=(1,0)),axes=(0,0))
        Koovv = np.tensordot(qoo, np.tensordot(gamma, qvv, axes=(1,0)),axes=(0,0))
        print "nocc = %s" % nocc
        print "nvirt = %s" % nvirt
        print "Kvvov.shape = %s" % str(Kvvov.shape)
        print "Kooov.shape = %s" % str(Kooov.shape)
        print "Kovoo.shape = %s" % str(Kovoo.shape)
        print "Koooo.shape = %s" % str(Koooo.shape)
        print "Koovv.shape = %s" % str(Koovv.shape)
#        exit(-1)
        # compute Q for occ-virt and virt-occ 
        Qov = np.zeros((nocc, nvirt))
        for i in range(0, nocc):
            for a in range(0, nvirt):
                for b in range(0, nvirt):
                    Qov[i,a] += XpY[I,i,b]*4*np.sum(Kvvov[a,b,:,:]*XpY) # (B12)
                Qov[i,a] += 4*np.sum(Kovvv[i,a,:,:]*Tvv[:,:]) # (B13)
                Qov[i,a] += 4*np.sum(Kovoo[i,a,:,:]*Too[:,:]) # (B13)
        Qvo = np.zeros((nocc, nvirt))
        for a in range(0, nvirt):
            for i in range(0, nocc):
                for j in range(0, nocc):
                    Qvo[i,a] += XpY[I,j,a]*4*np.sum(Kooov[j,i,:,:]*XpY) # (B15)
        # compute Q for occ-occ and virt-virt
        Qoo = np.zeros((nocc,nocc))
        for i in range(0, nocc):
            for j in range(0, nocc):
                # B(10)
                Qoo[i,j] += Omega[I]*(np.sum(XpY[I,i,:]*XmY[I,j,:])+np.sum(XmY[I,i,:]*XpY[I,j,:]))
                Qoo[i,j] -= np.sum(orbe[virt]*(XpY[I,i,:]*XpY[I,j,:]+XmY[I,i,:]*XmY[I,j,:]))
                Qoo[i,j] += 4*np.sum(Koooo[i,j,:,:]*Too[:,:])
                Qoo[i,j] += 4*np.sum(Koovv[i,j,:,:]*Tvv[:,:])
        Qvv = np.zeros((nvirt,nvirt))
        for a in range(0, nvirt):
            for b in range(0, nvirt):
                # B(17)
                Qvv[a,b] += Omega[I]*(np.sum(XpY[I,:,a]*XmY[I,:,b]) + np.sum(XmY[I,:,a]*XpY[I,:,b]))
                Qvv[a,b] += np.sum(orbe[occ]*(XpY[I,:,a]*XpY[I,:,b]+XmY[I,:,a]*XmY[I,:,b]))
        # right hand side (B19)
        R = Qov - Qvo
        # build (A+B)_(ia,jb)
        omega = np.zeros((nocc, nvirt))
        for i,occi in enumerate(occ):
            for a,virta in enumerate(virt):
                omega[i,a] = orbe[virta] - orbe[occi]
        omega = np.diag(np.reshape(omega, nocc*nvirt))
        omega = np.reshape(omega, (nocc,nvirt,nocc,nvirt))
        ApB = omega + 4*np.tensordot(qov, np.tensordot(gamma, qov, axes=(1,0)),axes=(0,0))
        # solve for Z
        ApB = np.reshape(ApB, (nocc*nvirt,nocc*nvirt))
        R = np.reshape(R, nocc*nvirt)
        Z = la.solve(ApB, -R)
        Z = np.reshape(Z,(nocc,nvirt))
        print "Z"
        print Z
        # find W
        # occ-occ
        Woo = np.zeros((nocc,nocc))
        for i in range(0, nocc):
            for j in range(0, nocc):
                Woo[i,j] = Qoo[i,j] + 4*np.sum(Kooov[i,j,:,:]*Z[:,:]) # B(20)
                if i == j:
                    Woo[i,j] /= 2.0
        # virt-virt
        Wvv = np.zeros((nvirt,nvirt))
        for a in range(0, nvirt):
            for b in range(0, nvirt):
                Wvv[a,b] = Qvv[a,b] # B(21)
                if a == b:
                    Wvv[a,b] /= 2.0
        # occ-virt
        Wov = np.zeros((nocc,nvirt))
        for i in range(0, nocc):
            for a in range(0, nvirt):
                Wov[i,a] = Qvo[i,a] + (orbe[virt[a]]-orbe[occ[i]])*Z[i,a] # B(22)
        W = np.bmat([[Woo,Wov],[zvo,Wvv]])
        # ASSEMBLE GRADIENT
        # gradients
        gradS, gradH0 = self.dftb.gradients_H0andS()
        # gradient of gamma-matrix
        gradG = self.dftb.grad_Gamma()

        # density matrix
        # D_(m,n) = sum_(j occ) C_(m,j) C(n,j)
        D = np.dot(orbs[:,occ], orbs[:,occ].transpose())
        # gradients of Mulliken charges
        gradQ = np.zeros((3*nat,nat))
        for C in range(0, nat):
            for p in range(0,3*nat):
                gradQ[p,C] = np.sum(D[at2orbs[C],:]*gradS[p,at2orbs[C],:])
        # compute grad H^(uxc)_(mn)
        gradHuxc = np.zeros(gradS.shape)
        for m in range(0, norb):
            for n in range(0,norb):
                # C(4)
                gAC = gamma[orbs2at[m],:]
                gBC = gamma[orbs2at[n],:]
                gradHuxc[:,m,n] += 0.5*gradS[:,m,n]*np.sum((gAC+gBC)*dq)
                gradHuxc[:,m,n] += 0.5*S[m,n]*np.dot(gradG[:,orbs2at[m],:]+gradG[:,orbs2at[n],:], dq)
                gradHuxc[:,m,n] += 0.5*S[m,n]*np.dot(gradQ, gAC+gBC)
        # P = T+Z
        Zov = Z
        Zoo = np.zeros((nocc,nocc))
        Zvv = np.zeros((nvirt,nvirt))
        Z = np.bmat([[Zoo,Zov],[zvo,Zvv]])
        P = T+Z
        # continue in atomic orbital basis
        Pao = np.dot(orbs, np.dot(P, orbs.transpose()))
        Wao = np.dot(orbs, np.dot(W, orbs.transpose()))
        # transform X+Y into AO basis
        XpYao = np.dot(orbs[:,occ],np.dot(XpY[I,:,:], orbs[:,virt].transpose()))
        # 
        gradK = np.zeros(3*nat)
        for A in range(0, nat):
            for m in at2orbs[A]:
                for k in at2orbs[A]:
                    for B in range(0, nat):
                        if A == B:
                            continue
                        for n in at2orbs[B]:
                            for l in at2orbs[B]:
                                gradK[:] += 0.25 * XpYao[m,n]*XpYao[k,l] * (\
  (S[k,l]*gradS[:,m,n] + S[m,n]*gradS[:,k,l])*(gamma[A,A]+gamma[A,B]+gamma[B,A]+gamma[B,B]) \
  +S[m,n]*S[k,l]*(gradG[:,A,A]+gradG[:,A,B]+gradG[:,B,A]+gradG[:,B,B]))

        gradE = np.zeros(3*nat)
        for p in range(0, 3*nat):
            gradE[p] += 2*np.sum(Pao*(gradH0[p,:,:]+gradHuxc[p,:,:]))
            gradE[p] -= np.sum(Wao*gradS[p,:,:])
            gradE[p] += 2*gradK[p]

        print "gradE"
        print gradE
        # compute gradients numerically
        from DFTB import XYZ, utils
        atomlist = self.dftb.atomlist
        x0 = XYZ.atomlist2vector(atomlist)
        gradEnum = utils.numerical_gradient(lambda x: self.tddftb.energy_func(x, self.dftb.atomlist, 0), x0)
        print "NUMERICAL GRADIENT"
        print gradEnum

        print "ANALYTICAL GRADIENT"
        print gradE

    def getExcGradients_wrong(self, I=0):
        """
        compute gradients of excitation energies
        I=0 is the first excited state, not the ground state!
        """
        XmY, XpY = self.tddftb.getXY()

        err = abs(np.sum(XmY[I,:,:]*XpY[I,:,:]) - 1.0)
        assert err < 1.0e-10, "err = %s" % err
        S, H0 = self.dftb.getH0andS()
        orbs = self.dftb.getKSCoefficients()
        orbe = self.dftb.getKSEnergies()
        gamma = self.dftb.getGamma()
        dq = self.dftb.getPartialCharges()
        qoo,qvv,qov = self.tddftb.getTransitionCharges()
        # LAGRANGE MULTIPLIERS Z and W
        occ, virt = self.tddftb.getActiveOrbitals()
        nocc, nvirt = len(occ), len(virt)
        norb = S.shape[0]
        nat,nat = gamma.shape
        at2orbs = self.dftb.getOrbitalsOnAtom()
        # orbs2at[p] gives the index of the atom on which orbital p sits
        orbs2at = np.zeros(norb,dtype=int)
        for A in range(0, nat):
            for m in at2orbs[A]:
                orbs2at[m] = A
        ###############################
        # solve for Z
        M = np.dot(orbs.transpose(), S)
        # occ-occ block
        boo = np.zeros((nocc,nocc))
        for r in range(0, nocc):
            for s in range(0, nocc):
                boo[r,s] = np.sum(XpY[I,r,:]*XpY[I,s,:] + (XmY[I,r,:]*XmY[I,s,:]))
        boo *= -0.5
        # virt-virt block
        bvv = np.zeros((nvirt,nvirt))
        for r in range(0, nvirt):
            for s in range(0, nvirt):
                bvv[r,s] = np.sum(XpY[I,:,r]*XpY[I,:,s] + (XmY[I,:,r]*XmY[I,:,s]))
        bvv *= 0.5
        
        zov = np.zeros((nocc,nvirt))
        zvo = np.zeros((nvirt,nocc))
        b = np.bmat([[boo,zov],[zvo,bvv]])
        Z = la.solve(M,b)
        Z = np.array(Z)
        ##############################
        # find W
        # Q_A[(X+Y)]  = sum_(i,a) (X+Y)_(ia) q_A^(ia)
        nat,nat = gamma.shape
        Q = np.zeros(nat)
        for A in range(0, nat):
            Q[A] = np.sum(XpY[I,:,:]*qov[A,:,:])
        # 
        print "Q"
        print Q
        # G_(a,b) = sum_J sum_I dqI gamma_IJ (delta(a on J) + delta(b on J))
        estatpot = np.dot(gamma, dq)
        G = np.zeros(S.shape)
        for J in range(0, nat):
            orbJ = at2orbs[J] # orbitals belonging to atom J
            G[orbJ,:] = estatpot[J]
            G[:,orbJ] += estatpot[J]
        # K_(l,k)
        # This is a very inefficient loop!!!!
        Koo = np.zeros((nocc, nocc))
        for l in range(0, nocc):
            for k in range(0, nocc):
                for A in range(0, nat):
                    for B in range(0, nat):
                        for a in range(0, nvirt):
                            Koo[l,k] += gamma[A,B]*XpY[I,k,a]*qov[A,l,a]*Q[B]
        Kvv = np.zeros((nvirt, nvirt))
        for l in range(0, nvirt):
            for k in range(0, nvirt):
                for A in range(0, nat):
                    for B in range(0, nat):
                        for i in range(0, nocc):
                            Koo[l,k] += gamma[A,B]*XpY[I,i,k]*qov[A,i,l]*Q[B]
        K = np.bmat([[Koo, zov],[zvo,Kvv]])
        K = np.array(K)
        # 
        en = np.diag(np.hstack((orbe[occ], orbe[virt])))
        # CHZ_(l,k) = sum_(a,b) C_(a,l) H^0_(a,b) Z_(b,k)
        CHZ = np.dot(orbs.transpose(), np.dot(H0, Z))
        # CSZ_(l,k) = sum_(a,b) C_(a,l) S_(a,b) Z_(b,k)
        CSZ = np.dot(orbs.transpose(), np.dot(S, Z))
        # CGSZ_(l,k) = sum_(a,b) C_(a,l) [G_(a,b) S_(a,b)] Z_(b,k)
        CGSZ = np.dot(orbs.transpose(), np.dot(G*S, Z))
        W = np.zeros((norb,norb))
        print "CHZ"
        print CHZ
        print "CSZ"
        print CSZ
        # occ-block
        for k in range(0, nocc):
            W[:,k] = CHZ[:,k] + 0.5*CGSZ[:,k]
        # occ and virt
        for k in range(0, norb):
            W[:,k] += -orbe[k] * CSZ[:,k]
        W += 2*K
        print "W"
        print W
        ##############################
        # modified density matrices
        # D_(a,b) = sum_(j occ) C_(a,j) Z(b,j)
        D = np.dot(orbs[:,occ], Z[:,occ].transpose())
        # energy weighted
        # enD_(a,b) = sum_(j occ) e
        enD = np.dot(orbe[occ]*orbs[:,occ], Z[:,occ].transpose())
        #
        # transform X+Y into AO basis
        XpYao = np.dot(orbs[:,occ],np.dot(XpY[I,:,:], orbs[:,virt].transpose()))
        # gammaQ[A] = sum_B gamma_AB QB[X+Y]
        gammaQ = np.dot(gamma, Q)
        # 
        dQS = np.zeros((norb,norb))
        XpY2T = XpYao + XpYao.transpose()
        for m in range(0, norb):
            for n in range(0, norb):
                dQS[m,n] += XpY2T[m,n]*gammaQ[orbs2at[m]]

        ###############################
        # assemble gradients
        # factors of dS/dp
        CWC = np.dot(orbs, np.dot(W, orbs.transpose()))
        Sfac = G*D - 2*enD - CWC + dQS
        # factors of dH/dp
        Hfac = 2*D
        # factors of dgamma/dp
        DS = D*S
        # DSsum_(B) = sum_(a,b) (delta(a on B) + delta(b on B))D_(a,b)*S_(a,b)
        DSsum = np.zeros(nat)
        for B in range(0, nat):
            orbB = at2orbs[B]
            DSsum[B] = np.sum(DS[at2orbs[B],:]) + np.sum(DS[:,at2orbs[B]])
        Gfac = np.zeros((nat,nat))    
        for A in range(0, nat):
            for B in range(0, nat):
                Gfac[A,B] += 2*Q[A]*Q[B] + dq[A] * DSsum[B]
        # gradients
        gradS, gradH0 = self.dftb.gradients_H0andS()
        # gradient of gamma-matrix
        gradG = self.dftb.grad_Gamma()

        gradE = np.zeros(3*nat)
        for p in range(0, 3*nat):
            gradE[p] += np.sum(Gfac*gradG[p,:,:])
            gradE[p] += np.sum(Sfac*gradS[p,:,:])
            gradE[p] += np.sum(Hfac*gradH0[p,:,:])

        print "GRAD"
        print gradE
        Fx = np.sum(gradE[0::3])
        Fy = np.sum(gradE[1::3])
        Fz = np.sum(gradE[2::3])
        print Fx
        print Fy
        print Fz
        
        # compute gradients numerically

        from DFTB import XYZ, utils
        atomlist = self.dftb.atomlist
        x0 = XYZ.atomlist2vector(atomlist)
        gradEnum = utils.numerical_gradient(lambda x: self.tddftb.energy_func(x, self.dftb.atomlist, 0), x0)
        print "NUMERICAL GRADIENT"
        print gradEnum

        print "ANALYTICAL GRADIENT"
        print gradE

        return gradE

            
