"""
gradients of excitation energies
"""
import numpy as np
import numpy.linalg as la
from scipy.special import erf 

from DFTB2 import gamma_function
from DFTB import XYZ, utils

def gradients_H0andS(atomlist, valorbs, SKT):
    """
    gradients of overlap matrix S and 0-order hamiltonian matrix H0
    using Slater-Koster Rules

    Parameters:
    ===========
    atomlist: list of tuples (Zi,[xi,yi,zi]) of atom types and positions
    valorbs: list of valence orbitals with quantum numbers (ni,li,mi)
    SKT: Slater Koster table
    
    Returns:
    ========
    grad S : gradS[i,m,n] is the derivative of S_mn with respect to the i-th coordinate
    grad H0: gradH0[i,m,n] is the derivative of H0_mn with respect to the i-th coordinate
    """
    Nat = len(atomlist)
    # count valence orbitals
    Norb = 0
    for i,(Zi,posi) in enumerate(atomlist):
        Norb += len(valorbs[Zi])
    gradS = np.zeros((3*Nat, Norb, Norb))
    gradH0 = np.zeros((3*Nat, Norb, Norb))
    # iterate over atoms
    mu = 0
    for i,(Zi,posi) in enumerate(atomlist):
        # iterate over orbitals on center i
        for (ni,li,mi) in valorbs[Zi]:
            # iterate over atoms
            nu = 0
            for j,(Zj,posj) in enumerate(atomlist):
                # iterate over orbitals on center j
                for (nj,lj,mj) in valorbs[Zj]:
                    if mu == nu:
                        pass
                    else:
                        if Zi <= Zj:
                            # the first atom given to getHamiltonian() or getOverlap()
                            # has to be always the one with lower atomic number
                            if i == j:
                                # different orbitals on the same atom
                                # no contribution
                                assert mu != nu
                                s_deriv = np.zeros(3)
                                h0_deriv = np.zeros(3)
                            else: 
                                # the hardcoded Slater-Koster rules compute the gradient 
                                # with respect to r = posj - posi
                                # but we want the gradient with respect to posi, so an additional
                                # minus sign is introduced
                                s_deriv  = -np.array(SKT[(Zi,Zj)].getOverlap(li,mi,posi, lj,mj,posj, deriv=1))
                                h0_deriv = -np.array(SKT[(Zi,Zj)].getHamiltonian0(li,mi,posi, lj,mj,posj, deriv=1))
                        else:
                            # swap atoms if Zj > Zi, since posi and posj are swapped, the gradient
                            # with respect to r = posi - posj equals the gradient with respect to
                            # posi, so no additional minus sign is needed.
                            s_deriv  = np.array(SKT[(Zj,Zi)].getOverlap(lj,mj,posj, li,mi,posi, deriv=1))
                            h0_deriv = np.array(SKT[(Zj,Zi)].getHamiltonian0(lj,mj,posj, li,mi,posi, deriv=1))
                            #
                        gradS[3*i:3*(i+1),mu,nu] += s_deriv
                        gradH0[3*i:3*(i+1),mu,nu] += h0_deriv
                        # S and H0 are hermitian/symmetric
                        gradS[3*i:3*(i+1),nu,mu] += s_deriv
                        gradH0[3*i:3*(i+1),nu,mu] += h0_deriv

                    nu += 1
            mu += 1
    return gradS, gradH0

class GammaFunction: # base class
    """
    gamma_AB = int F_A(r-RA) * 1/|RA-RB| * F_B(r-RB) d^3r
    """
    def __init__(self, hubbard_U):
        """
        Parameters:
        ===========
        hubbard_U: dictionary hubbard_U[Z] should give the U parameter belonging
          to an atom with atomic number Z
        """
        self.hubbard_U = hubbard_U
    def g(self, R, ZA, ZB):
        """evaluate gamma(R)"""
        pass
    def g_deriv(self, R, ZA, ZB):
        """evaluate the derivative d/dR gamma(R)"""
        pass
    def g_limit0(self, Z):
        """evaluate the limit R->0 of gamma(R)"""
        return self.hubbard_U[Z]

class GammaGaussian(GammaFunction):
    """
    spherical charge fluctuations are modelled as Gaussians 

       FA(|r-RA|) = 1/(2 pi sigmaA^2)^(3/2) * exp( - (r-RA)^2/(2*sigmaA^2) )

    """
    def __init__(self, hubbard_U):
        """construct the C_AB matrix"""
        self.hubbard_U = hubbard_U
        sigma = {}
        for Z,U in hubbard_U.iteritems():
            sigma[Z] = 1.0/(np.sqrt(np.pi)*U)
        # 
        self.C = {}
        Zs = hubbard_U.keys()
        for ZA in Zs:
            for ZB in Zs:
                self.C[(ZA,ZB)] = 1.0/np.sqrt(2*(sigma[ZA]**2 + sigma[ZB]**2))
    def g(self, R, ZA, ZB):
        assert R > 0.0
        gAB = erf(self.C[(ZA,ZB)]*R)/R
        return gAB
    def g_deriv(self, R, ZA, ZB):
        assert R > 0.0
        c = self.C[(ZA,ZB)]
        gd = 2.0*c/np.sqrt(np.pi) * np.exp(-(c*R)**2)/R - erf(c*R)/R**2
        return gd

class GammaSlater(GammaFunction):
    """
    spherical charge fluctuation around an atom A are modelled as Slater functions

      FA(|r-RA|) = tA^3/(8 pi)*exp(-tA*|r-RA|)

    """
    def __init__(self, hubbard_U):
        self.hubbard_U = hubbard_U
        self.tau = {}
        for Z,U in hubbard_U:
            self.tau[Z] = 16.0/5.0 * U
    def g(self, R, ZA,ZB):
        assert R > 0.0
        tA = self.tau[ZA]
        tB = self.tau[ZB]
        tAR = tA*R
        tBR = tB*R
        tA2 = tA**2
        tB2 = tB**2
        denom = 2*(tA2-tB2)**3
        fA = tB2**2*(tB2*(2+tAR) - tA2*(6+tAR))/denom
        fB = tA2**2*(tA2*(2+tBR) - tB2*(6+tAR))/denom
        gAB = 1.0/R * (1 + fA*np.exp(-tAR) - fB*np.exp(-tBR))
        return gAB
    def g_deriv(self, R, ZA,ZB):
        assert R > 0.0
        tA = self.tau[ZA]
        tB = self.tau[ZB]
        tAR = tA*R
        tBR = tB*R
        tA2 = tA**2
        tB2 = tB**2
        denom = 2*(tA2-tB2)**3
        fB = (2+tBR*(2+tBR))*tA2 - (6+tBR*(6+tBR))*tB2
        fA = (2+tAR*(2+tAR))*tB2 - (6+tAR*(6+tAR))*tA2
        gd = 1.0/R**2 * (-1 \
              + 1.0/denom * (tA2**2*fB*np.exp(-tBR) - tB2**2*fA*np.exp(-tAR)))
        return gd

class SwitchingFunction: # base class
    def f(self, R):
        pass
    def f_deriv(self, R):
        pass

class ErfSwitching(SwitchingFunction):
    """ f(R) = erf(R/Rlr)"""
    def __init__(self, Rlr):
        self.Rlr = Rlr
    def f(self, R):
        sw = erf(R/self.Rlr)
        return sw
    def f_deriv(self, R):
        dsw = 2.0/(np.sqrt(np.pi)*self.Rlr) * np.exp(-(R/self.Rlr)**2)
        return dsw

class NoSwitching(SwitchingFunction):
    """ f(R) = 0"""
    def f(self, R):
        return 0
    def f_deriv(self, R):
        return 0

class GammaMatrix:
    def __init__(self, gamma_func, switching_func):
        self.gamma_func = gamma_func
        self.switching_func = switching_func
    def distance_matrix(self, atomlist):
        Nat = len(atomlist)
        distances = np.zeros((Nat,Nat))
        directions = np.zeros((Nat,Nat,3))
        for i,(Zi,posi) in enumerate(atomlist):
            for j,(Zj,posj) in enumerate(atomlist):
                if i == j:
                    distances[i,j] = 0.0
                    directions[i,j,:] = np.zeros(3)
                elif i < j:
                    R = np.array(posi) - np.array(posj)
                    Rij = la.norm(R)
                    eij = R/Rij
                    distances[i,j] = Rij
                    directions[i,j,:] = eij
                else:
                    distances[i,j] = distances[j,i]
                    directions[i,j,:] = -directions[j,i]
        return distances, directions

    def matrices_pairwise(self, atomlist, distances):
        """build the gamma-matrices and its gradients"""
        Nat = len(atomlist)
        g0 = np.zeros((Nat,Nat))  # gamma
        g1 = np.zeros((Nat,Nat))  # derivative of gamma
        g0lr = np.zeros((Nat,Nat)) # long-range gamma
        g1lr = np.zeros((Nat,Nat)) # derivative of long-range gamma
        for i,(Zi,posi) in enumerate(atomlist):
            for j,(Zj,posj) in enumerate(atomlist):
                if i == j:
                    g0[i,j] = self.gamma_func.g_limit0(Zi)
                    g1[i,j] = 0.0
                    g0lr[i,j] = 0.0
                    g1lr[i,j] = 0.0
                elif i < j:
                    Rij = distances[i,j]
                    g0[i,j] = self.gamma_func.g(Rij, Zi, Zj)
                    g1[i,j] = self.gamma_func.g_deriv(Rij, Zi, Zj)
                    g0lr[i,j] = self.switching_func.f(Rij) * g0[i,j]
                    g1lr[i,j] = self.switching_func.f_deriv(Rij) * g0[i,j] \
                               +self.switching_func.f(Rij) * g1[i,j]
                else:
                    g0[i,j] = g0[j,i]
                    g1[i,j] = g1[j,i]
                    g0lr[i,j] = g0lr[j,i]
                    g1lr[i,j] = g1lr[j,i]
        return g0, g1, g0lr, g1lr
    def matrices_AObasis(self, atomlist, valorbs):
        """
        form gamma matrices in AO basis
          gamma_ab = sum_A sum_B delta(a in A) delta(b in B) gamma_AB
        """
        distances, directions = self.distance_matrix(atomlist)
        g0,g1,g0lr,g1lr = self.matrices_pairwise(atomlist, distances)
        Nat = len(atomlist)
        # count valence orbitals
        Norb = 0
        for i,(Zi,posi) in enumerate(atomlist):
            Norb += len(valorbs[Zi])

        g0_AO = np.zeros((Norb,Norb))
        g1_AO = np.zeros((3*Nat,Norb,Norb))
        g0lr_AO = np.zeros((Norb,Norb))
        g1lr_AO = np.zeros((3*Nat,Norb,Norb))

        # iterate over atoms
        mu = 0
        for i,(Zi,posi) in enumerate(atomlist):
            # iterate over orbitals on center i
            for (ni,li,mi) in valorbs[Zi]:
                # iterate over atoms
                nu = 0
                for j,(Zj,posj) in enumerate(atomlist):
                    # iterate over orbitals on center j
                    for (nj,lj,mj) in valorbs[Zj]:
                        g0_AO[mu,nu] = g0[i,j]
                        g0_AO[nu,mu] = g0[i,j]
                        g0lr_AO[mu,nu] = g0lr[i,j]
                        g0lr_AO[nu,mu] = g0lr[i,j]
                        if i != j:
                            eij = directions[i,j,:]
                            #
                            g1_AO[3*i:3*i+3,mu,nu] += g1[i,j] * eij
                            g1_AO[3*i:3*i+3,nu,mu] += g1[i,j] * eij
                            #
                            g1lr_AO[3*i:3*i+3,mu,nu] += g1lr[i,j] * eij
                            g1lr_AO[3*i:3*i+3,nu,mu] += g1lr[i,j] * eij

                        nu += 1
                mu += 1
        g1_AT = np.zeros((3*Nat,Nat,Nat))
        g1lr_AT = np.zeros((3*Nat,Nat,Nat))
        for i in range(0, Nat):
            for j in range(0, Nat):
                eij = directions[i,j,:]
                g1_AT[3*i:3*i+3,i,j] = g1[i,j] * eij
                g1lr_AT[3*i:3*i+3,i,j] = g1lr[i,j] * eij
                
        return g0, g1_AT, g0lr, g1lr_AT, g0_AO, g1_AO, g0lr_AO, g1lr_AO

class Gradients:
    def __init__(self, tddftb):
        self.tddftb = tddftb
        self.dftb = tddftb.dftb2
        self.gf = GammaGaussian(self.dftb.hubbard_U)
        if self.dftb.long_range_correction == 1:
            self.sw = ErfSwitching(self.dftb.long_range_radius)
        else:
            self.sw = NoSwitching()
        self.gm = GammaMatrix(self.gf, self.sw)
    def activeOrbitals(self):
        orbe = self.dftb.getKSEnergies()
        orbs = self.dftb.getKSCoefficients()
        occ, virt = self.tddftb.getActiveOrbitals()
        active = np.hstack((occ,virt))
        nocc = len(occ)
        nvirt = len(virt)
        orbe_occ = orbe[occ]
        orbe_virt = orbe[virt]
        orbs_occ = orbs[:,occ]
        orbs_virt = orbs[:,virt]
        return orbe_occ, orbe_virt, orbs_occ, orbs_virt, nocc, nvirt
    def density_matrix_ref(self):
        """
        construct reference density matrix D0
        """
        P0diag = []
        # iterate over atoms
        for i,(Zi,posi) in enumerate(self.dftb.atomlist):
            # iterate over orbitals on center i
            for iv,(ni,li,mi) in enumerate(self.dftb.valorbs[Zi]):
                # how many valence electrons are put into the nl-shell?
                P0diag.append( self.dftb.valorbs_occupation[Zi][iv] )
        print "reference density matrix"
        print P0diag
        P0 = np.diag(P0diag)
        return P0
    def gradient(self, I=0):
        """
        solve for the Lagrange multipliers Z and W
        and build gradient
        """
        XmY, XpY = self.tddftb.getXY()
        Omega = self.tddftb.getExcEnergies()

        # select state I
        XmY = XmY[I,:,:]
        XpY = XpY[I,:,:]
        Omega = Omega[I]
        
        orbe_occ, orbe_virt, orbs_occ, orbs_virt, nocc,nvirt = self.activeOrbitals()
        Norb = nocc+nvirt
        qoo,qvv,qov = self.tddftb.getTransitionCharges()

        Uab = np.zeros((nvirt,nvirt))
        Uij = np.zeros((nocc,nocc))
        Vab = np.zeros((nvirt,nvirt))
        Vij = np.zeros((nocc,nocc))
        Kab = np.zeros((nvirt,nvirt))
        Kij = np.zeros((nocc,nocc))
        # T vectors
        Tab = np.zeros((nvirt,nvirt))
        Tij = np.zeros((nocc,nocc))

        # vectors Uab, Vab, Kab, Tab, virt-virt block
        for a in range(0, nvirt):
            for b in range(0, nvirt):
                for i in range(0, nocc):
                    Uab[a,b] += XpY[i,a]*XmY[i,b] + XmY[i,a]*XpY[i,b]
                    Vab[a,b] += orbe_occ[i]*(XpY[i,a]*XpY[i,b] + XmY[i,a]*XmY[i,b])
                    Kab[a,b] += XpY[i,a]*XpY[i,b]
                    Tab[a,b] += 0.5*(XpY[i,a]*XpY[i,b]+XmY[i,a]*XmY[i,b])
        # and vectors  and Uij, Vij, Kij, Tij, occ-occ block
        for i in range(0, nocc):
            for j in range(0, nocc):
                for a in range(0, nvirt):
                    Uij[i,j] += XpY[i,a]*XmY[j,a] + XmY[i,a]*XpY[j,a]
                    Vij[i,j] += orbe_virt[a]*(XpY[i,a]*XpY[j,a] + XmY[i,a]*XmY[j,a])
                    Kij[i,j] += XpY[i,a]*XpY[j,a]
                    Tij[i,j] += 0.5*(XpY[i,a]*XpY[j,a]+XmY[i,a]*XmY[j,a])
        # Gamma matrices
        g0, g1, g0lr, g1lr, g0_AO, g1_AO, g0lr_AO, g1lr_AO = self.gm.matrices_AObasis(self.dftb.getGeometry(), self.dftb.valorbs)

        Nat, Nat = g0.shape
        # sum_B g_AB q_B^(pq)
        gq_oo = np.tensordot(g0, qoo, axes=(1,0))
        gq_ov = np.tensordot(g0, qov, axes=(1,0))
        gq_lr_oo = np.tensordot(g0lr, qoo, axes=(1,0))
        gq_lr_ov = np.tensordot(g0lr, qov, axes=(1,0))
        gq_lr_vv = np.tensordot(g0lr, qvv, axes=(1,0))
        # 2 * sum_ab (q_A^(ab) K_ab) - sum_ij (q_A^(ij) K_ij)
        qK = np.zeros(Nat)
        for A in range(0, Nat):
            qK[A] = 2 * (np.sum(qvv[A,:,:]*Kab) - np.sum(qoo[A,:,:]*Kij))
        # Gia
        Gia = np.tensordot(qK, gq_ov, axes=(0,0))
        tmp = np.tensordot(Kab, gq_lr_vv, axes=(1,2))
        tmp2 = np.tensordot(qov, tmp, axes=(2,0))
        tmp3 = np.sum(np.sum(tmp2, axis=2), axis=0)
        Gia -= tmp3
        tmp = np.tensordot(Kij, gq_lr_ov, axes=(1,1))
        tmp2 = np.tensordot(qoo, tmp, axes=(1,0))
        tmp3 = np.sum(np.sum(tmp2, axis=2), axis=0)
        Gia += tmp3
        print "Gia"
        print Gia
        # Gij
        Gij = np.tensordot(qK, gq_oo, axes=(0,0))
        tmp = np.tensordot(Kab, gq_lr_ov, axes=(1,2))
        tmp2 = np.tensordot(qov, tmp, axes=(2,0))
        tmp3 = np.sum(np.sum(tmp2, axis=2), axis=0)
        Gij -= tmp3
        tmp = np.tensordot(Kij, gq_lr_oo, axes=(1,1))
        tmp2 = np.tensordot(qoo, tmp, axes=(1,0))
        tmp3 = np.sum(np.sum(tmp2, axis=2), axis=0)
        Gij += tmp3
        print "Gij"
        print Gij
        # build Q
        Qab = Omega * Uab + Vab
        Qij = Omega * Uij - Vij + Gij
        Qia = Gia
        Qai = 0.0
        
        # right hand side
        Ria = -Qia
        # build (A+B)_(ia,jb)
        omega = np.zeros((nocc,nvirt))
        for i in range(0, nocc):
            for a in range(0, nvirt):
                omega[i,a] = orbe_virt[a] - orbe_occ[i]
        #
        ApB = np.reshape(np.diag(np.reshape(omega, nocc*nvirt)), (nocc,nvirt,nocc,nvirt))
        ApB += 4*np.tensordot(qov, gq_ov, axes=(0,0))
        tmp = np.tensordot(qoo, gq_lr_vv, axes=(0,0))
        ApB -= np.swapaxes(tmp, 1, 2)
        tmp = np.tensordot(qov, gq_lr_ov, axes=(0,0))
        ApB -= np.swapaxes(tmp, 1, 3)

        ApB = np.reshape(ApB, (nocc*nvirt,nocc*nvirt))
        Ria = np.reshape(Ria, nocc*nvirt)
        # solve for Z
        Z = la.solve(ApB, Ria)
        Zia = np.reshape(Z, (nocc,nvirt))
        # build W
        # sum_(jb) (A+B)_(ia,jb) Z_jb
        qZ = np.zeros(Nat)
        for B in range(0, Nat):
            qZ[B] = np.sum(qov[B,:,:]*Zia)
        tmp = np.dot(g0, qZ)
        ApBZ = 4*np.tensordot(qoo, tmp, axes=(0,0))
        tmp = np.tensordot(gq_lr_oo, Zia, axes=(2,0))
        tmp2 = np.tensordot(qov, tmp, axes=(2,2))
        tmp3 = np.sum(np.sum(tmp2, axis=2), axis=0)
        ApBZ -= tmp3
        tmp = np.tensordot(gq_lr_ov, Zia, axes=(2,1))
        tmp2 = np.tensordot(qoo, tmp, axes=(2,2))
        tmp3 = np.sum(np.sum(tmp2, axis=2), axis=0)
        ApBZ -= tmp3

        Wij = Qij + ApBZ
        for i in range(0, nocc):
            Wij[i,j] /= 2.0
        Wia = np.zeros((nocc,nvirt))
        for i in range(0, nocc):
            Wia[i,:] = Zia[i,:]*orbe_occ[i]
        Wai = Wia.transpose()
        Wab = Qab
        for a in range(0, nvirt):
            Wab[a,a] /= 2.0
        W = np.bmat([[Wij, Wia], [Wai, Wab]])
        W = np.array(W)

        # ASSEMBLE GRADIENTS
        gradS, gradH0 = gradients_H0andS(self.dftb.getGeometry(), self.dftb.valorbs, 
                                         self.dftb.SKT)
        S = self.dftb.S
        # linear operators
        def F(v):
            vp = v+v.transpose()
            Sv = np.sum(S*vp, axis=1)
            gSv = np.dot(g0_AO, Sv)
            gdSv = np.zeros((3*Nat,Norb))
            dgSv = np.zeros((3*Nat,Norb))
            for nc in range(0, 3*Nat):
                gdSv[nc,:] = np.dot(g0_AO,np.sum(gradS[nc,:,:]*vp, axis=1))
                dgSv[nc,:] = np.dot(g1_AO[nc,:,:], Sv)
            f = np.zeros((3*Nat,Norb,Norb))
            # a and b are AOs
            for a in range(0, Norb):
                for b in range(0, Norb):
                    f[:,a,b] = \
                        gradS[:,a,b]*(gSv[a] + gSv[b]) \
                       +S[a,b]*(dgSv[:,a]+dgSv[:,b]) \
                       +S[a,b]*(gdSv[:,a]+gdSv[:,b])
            f *= 0.25
            return f
        def Flr(v):
            Sv = np.dot(S, v)
            SvT = np.dot(S, v.transpose())
            gv = g0lr_AO * v
            dSv24 = np.tensordot(gradS, v, axes=(2,1))
            dSv23 = np.tensordot(gradS, v, axes=(2,0))
            dgv = np.zeros((3*Nat,Norb,Norb))
            for nc in range(0,3*Nat):
                dgv[nc,:,:] = g1lr_AO[nc,:,:]*v
            #
            flr = np.zeros((3*Nat,Norb,Norb))
            # 1st term in eqn. for F_ab^lr
            tmp1 = np.tensordot(gradS, Sv, axes=(2,1))
            for nc in range(0, 3*Nat):
                flr[nc,:,:] += g0lr_AO*tmp1[nc,:,:]
            # 2nd term
            tmp2 = np.zeros((3*Nat,Norb,Norb))
            for nc in range(0, 3*Nat):
                tmp2[nc,:,:] = dSv24[nc,:,:]*g0lr_AO
            flr += np.tensordot(tmp2, S, axes=(2,1))
            # 3rd term
            flr += np.tensordot(gradS, Sv*g0lr_AO, axes=(2,1))
            # 4th term
            flr += np.tensordot(gradS, np.dot(S, gv), axes=(2,1))
            # 5th term
            tmp5 = np.swapaxes(np.tensordot(S, dSv23, axes=(1,2)), 0,1)
            for nc in range(0, 3*Nat):
                flr[nc,:,:] += g0lr_AO*tmp5[nc,:,:]
            # 6th term
            flr += np.swapaxes(np.tensordot(SvT*g0lr_AO, gradS, axes=(1,2)), 0,1)
            # 7th term
            tmp7 = np.zeros((3*Nat,Norb,Norb))
            for nc in range(0, 3*Nat):
                tmp7[nc,:,:] = dSv23[nc,:,:]*g0lr_AO
            flr += np.swapaxes(np.tensordot(S, tmp7, axes=(1,2)), 0,1)
            # 8th term
            tmp8 = np.tensordot(gradS, gv, axes=(2,0))
            flr += np.swapaxes(np.tensordot(S, tmp8, axes=(1,2)), 0,1)
            # 9th term
            tmp9 = np.dot(S, Sv.transpose())
            for nc in range(0, 3*Nat):
                flr[nc,:,:] += g1lr_AO[nc,:,:] * tmp9
            # 10th term
            tmp10 = np.zeros((3*Nat,Norb,Norb))
            for nc in range(0, 3*Nat):
                tmp10[nc,:,:] = SvT * g1lr_AO[nc,:,:]
            flr += np.tensordot(tmp10, S, axes=(2,1))
            # 11th term
            tmp11 = np.zeros((3*Nat,Norb,Norb))
            for nc in range(0, 3*Nat):
                tmp11[nc,:,:] = Sv * g1lr_AO[nc,:,:]
            flr += np.swapaxes(np.tensordot(S, tmp11, axes=(1,2)), 0,1)
            # 12th term
            tmp12 = np.tensordot(dgv, S, axes=(1,1))
            flr += np.swapaxes(np.tensordot(S, tmp12, axes=(1,1)), 0,1)
            ###
            flr *= 0.25
            return flr
        # density matrix
        D = 2*np.dot(orbs_occ, orbs_occ.transpose())
        # reference density matrix
        D0 = self.density_matrix_ref()
        ##############################
        # dH/dR
        FDmD0 = F(D-D0)

        FlrD = Flr(D)
        gradH = gradH0 + FDmD0 - 0.5 * FlrD

        # transform vectors to AO basis
        Too = np.dot(orbs_occ,np.dot(Tij,orbs_occ.transpose()))
        Tvv = np.dot(orbs_virt,np.dot(Tab,orbs_virt.transpose()))
        Zao = np.dot(orbs_occ,np.dot(Zia,orbs_virt.transpose()))
        orbs = np.hstack((orbs_occ,orbs_virt))
        Wao = np.dot(orbs,np.dot(np.triu(W), orbs.transpose()))
        XpYao = np.dot(orbs_occ,np.dot(XpY,orbs_virt.transpose()))
        XmYao = np.dot(orbs_occ,np.dot(XmY,orbs_virt.transpose()))

        gradExc = np.zeros(3*Nat)
        f = F(XpYao)
        flr_p = Flr(XpYao+XpYao.transpose())
        flr_m = Flr(XmYao-XmYao.transpose())
        for nc in range(0, 3*Nat):
            gradExc[nc] = np.sum(\
                  gradH[nc,:,:]*(Tvv-Too+Zao)) - np.sum(gradS[nc,:,:]*Wao) \
                + 2*np.sum(XpYao*f[nc,:,:]) \
                - 0.5*np.sum(XpYao*flr_p[nc,:,:]) \
                - 0.5*np.sum(XmYao*flr_m[nc,:,:])

        # GRADIENT OF GROUND STATE ENERGY
        # energy weighted density matrix
        en_orbs = np.copy(orbs_occ)
        for i in range(0, nocc):
            en_orbs[:,i] *= orbe_occ[i]
        Den = 2*np.dot(en_orbs, orbs_occ.transpose())
        
        gradE0 = np.zeros(3*Nat)
        for nc in range(0, 3*Nat):
            gradE0[nc] = np.sum(  gradH0[nc,:,:]*D \
                                + 0.5*FDmD0[nc,:,:]*(D-D0) \
                                - 0.25*FlrD[nc,:,:]*D \
                                - gradS[nc,:,:]*Den)
            
        #
        return gradE0, gradExc

    def getExcGradients(self, I=0):
        """
        compute gradients of excitation energies
        I=0 is the first excited state, not the ground state!

        according to "Analytical Excited State Forces for the TD-DFTB Method"
        by Heringer et al., J. Comput. Chem., (2007), 28, 16. 
        """
        gradE0, gradExc = self.gradient(I=I)
        # 
        # compute gradients numerically
        atomlist = self.dftb.atomlist
        x0 = XYZ.atomlist2vector(atomlist)

        """
        # ground state
        gradE0Num = utils.numerical_gradient(lambda x: self.dftb._energy_func(x, self.dftb.atomlist, "Eelec"), x0)
        print "ANALYTICAL GRADIENT"
        print gradE0

        print "NUMERICAL GRADIENT"
        print gradE0Num
        """


        # excited states
        gradExcNum = utils.numerical_gradient(lambda x: self.tddftb.energy_func(x, self.dftb.atomlist, 0), x0)
        print "ANALYTICAL GRADIENT"
        print gradExc

        print "NUMERICAL GRADIENT"
        print gradExcNum




            
