"""
Implementation of the DREIDING force field for QM/MM calculations with periodic boundary conditions
in molecular crystals.
"""
import numpy as np

from DFTB import XYZ, AtomicData

# class that wraps C-code
class PeriodicForceField:
    def __init__(self, atomlist, atomtypes, lattice_vectors, verbose=1):
        """
        Builds a forces field with periodic boundary conditions.
        valence terms:
          bond stretching, angle bending, torsion and spectroscopic inversion
        non-bonding terms:
          van der Waals (Lennard Jones potential)
        The Coulomb interaction is not included, so all atoms should be approximately neutral. 
        The parameters are taken from the DREIDING force field. 

        So far only two atom-types exist:
          0: hydrogen 
          1: aromatic carbon

        Parameters:
        ===========
        atomlist: geometry which defined the bonding topology, 
          list of tuples (Zat,[x,y,z]) for each atom
        atomtypes: list with indeces of atom types
        lattice_vectors: 3 lists with translation vectors of lattice
          [[ax,ay,az],[bx,by,bz],[cx,cy,cz]]
        verbose: print additional information about the constructed force field
        """
        import ff
        self.force_field = ff.ForceField(atomlist, atomtypes, lattice_vectors, verbose=verbose)
    def getEnergyAndGradient(self, coords):
        """
        Computes the energy and gradient on the atoms in the central unit cell.
        The number and order of atoms should remain the same as used when defining the force field.

        Parameters:
        ===========
        coords: numpy array with shape (3*Nat), coords[3*i:3*(i+1)] are the x,y,z coordinates 
           of the i-th atom in bohr

        Returns:
        ========
        energy: total force field energy in Hartree
        grads: numpy array with shape (3*Nat), grads[3*i:3*(i+1)] is the gradient on the i-th atom in Hartree/bohr
        """
        energy, grad = self.force_field.getEnergyAndGradient(coords)
        return energy, grad
    # The following methods are only defined to make the interface compatible with Gaussian.UFF_handler
    def calc(self, atomlist, **kwds):
        coords = XYZ.atomlist2vector(atomlist)
        self.energy, self.grad = self.getEnergyAndGradient(coords)
    def get_MM_Energy(self):
        return self.energy
    def get_MM_Gradient(self):
        return self.grad
    
    
def read_force_field(ff_file, units="Angstrom", fragment_id=AtomicData.atomic_number):
    """
    A force field file is structured like an xyz-file.
    1st line: number of atoms Nat
    2nd line: comment (line must not be empty)
    lines 3 to Nat+3: element name, x,y,z and atom type
       e.g.:      C   0.0 0.0 0.0    1
    The next lines are optional and contain the lattice vectors defining
    the unit cell starting with the symbol TV (translation vector):
       e.g.       Tv  1.0 0.0 0.0 
                  Tv  0.0 1.0 0.0
                  Tv  0.0 0.0 1.0


    Returns:
    ========
    atomlist: list of tuples (Zati,[xi,yi,zi])
    atomtypes: list of atom types for each atom
    lattice_vectors: list of at most 3 vectors defining the unit cell
    """
    fh = open(ff_file)
    line = fh.readline()
    # read number of atoms
    nat = int(line.split()[0])
    # skip commenct
    fh.readline()
    #
    atomlist = []
    atomtypes = []
    lattice_vectors = []
    for i in xrange(nat+3):
        # read nat atoms and at most 3 lattice vectors
        line = fh.readline()
        if not line:
            # end of file reached
            break
        words = line.split()
        x,y,z = map(float, words[1:4])
        if units == "Angstrom":
            x,y,z = map(lambda c: c/AtomicData.bohr_to_angs, [x,y,z])
        if words[0] == "Tv":
            lattice_vectors.append( [x,y,z] )
            continue
        atno = fragment_id(words[0])
        atomlist.append((atno, [x,y,z]))
        atomtypes.append(int(words[4]))
    fh.close()
    assert len(lattice_vectors) == 3, "Need 3 lattice vectors, got %d!" % len(lattice_vectors)
    return atomlist, atomtypes, lattice_vectors

def enlarged_unitcell(atomlist, lattice_vectors, nmax=1):
    """
    add neighbouring n cells to geometry
    """
    if nmax == 0:
        # only unitcell
        return atomlist
    a1 = np.array(lattice_vectors[0])
    a2 = np.array(lattice_vectors[1])
    a3 = np.array(lattice_vectors[2])

    atomlist_enlarged = []
    ns = range(-nmax,nmax+1)
    for n1 in ns:
        for n2 in ns:
            for n3 in ns:
                for Zi,posi in atomlist:
                    tvec = n1*a1 + n2*a2 + n3*a3
                    posi_translated = np.array(posi) + tvec
                    atomlist_enlarged.append( (Zi, posi_translated) )
    return atomlist_enlarged
                    
if __name__ == "__main__":
    import numpy as np
    from DFTB import XYZ, utils
    import sys

    if len(sys.argv) < 2:
        print "Usage: python %s <.ff file>" % sys.argv[0]
        print "  constructs force field and checks that numerical and analytical gradients agree"
        exit(-1)

    ff_file = sys.argv[1]  #"h2.ff" #"ethene.ff" #"pyrene_crystal_expanded.ff" #

    atomlist, atomtypes, lattice_vectors = read_force_field(ff_file)
    pff = PeriodicForceField(atomlist, atomtypes, lattice_vectors)

    coords = XYZ.atomlist2vector(atomlist)

    def energy_func(x):
        energy, grad = pff.getEnergyAndGradient(x)
        return energy

    grad_num = utils.numerical_gradient(energy_func, coords, 1.0e-4)
    energy, grad_analytic = pff.getEnergyAndGradient(coords)

    Nat = len(atomlist)
    print "numerical gradient"
    for i in range(0, Nat):
        print grad_num[3*i:3*(i+1)];
    print "analytical gradient"
    for i in range(0, Nat):
        print grad_analytic[3*i:3*(i+1)]

    grad_dif = grad_num - grad_analytic
    print "difference"
    for i in range(0, Nat):
        print grad_dif[3*i:3*(i+1)]

    print "error: %s" % np.sum(abs(grad_dif))

