#!/usr/bin/env python
"""
Optimizing large zn-porphyrin tapes is not feasible because of the memory requirements for calculating
analytical gradients. Optimizing short tapes shows that the lengths of the fused bonds are not uniform
along the tape: The fused bonds in the center of the tapes are shorter than at the edges. Therefore the
optimization will be restricted to two parameters, which model the relative stretching
as a function of z. The z-coordinate of each atom is stretched by a factor S,
  z  ->  S_n(z;u,v)*z
The factor S depends on the parameters u and v and the coordinate z in the following way
  S_n(u,v) = (u - v*cos(2*n*pi*z/L))
L is the length of the tape. The center of the tape is located in z=0, the edges are at z=-L/2 and z=+L/2
The parameter u adjusts the average bond length and the parameter v is responsible for the bond length variation.
The optimization starts with u=1 and v=0. The integer n 
"""
import numpy as np
from scipy import optimize

from DFTB import utils, XYZ
from DFTB.DFTB2 import DFTB2
from DFTB.Modeling import MolecularCoords as MolCo

import sys
import os

if __name__ == "__main__":
    usage =  "python %s <xyz-file with porphyrin tape> <xyz output file>" % sys.argv[0]
    usage += "Type --help to show all options for DFTB"
    parser = utils.OptionParserFuncWrapper(
        [DFTB2.__init__, 
         DFTB2.runSCC], usage)
    options, args = parser.parse_args(DFTB2.__init__)
    if len(args) < 2:
        print usage
        exit(-1)
    xyz_file = args[0]
    opt_file = args[1]
    
    atomlist = XYZ.read_xyz(xyz_file)[0]
    atomlist_std = MolCo.standard_orientation(atomlist)
    vec = XYZ.atomlist2vector(atomlist_std)
    # length of the tape
    L = vec[2::3].max() - vec[2::3].min()
    print "Length of tape: %s" % L
    n = 1
    
    dftb2 = DFTB2(atomlist_std, **options)
    scf_options, args = parser.parse_args(dftb2.runSCC)

    try:
        os.remove(opt_file)
    except OSError:
        pass
    def energy(uv):
        u,v = uv[0], uv[1]
        # scale z-coordinate by squeezing factor S
        vec_squeezed = np.copy(vec)
        z = vec[2::3]
        assert len(z) == len(atomlist)
        # 
        S = (u - v*np.cos(2*n*np.pi*z/L))
        # distort geometry along z-axis
        vec_squeezed[2::3] = S*z
        atomlist_squeezed = XYZ.vector2atomlist(vec_squeezed, atomlist_std)
        dftb2.setGeometry(atomlist_squeezed)
        en = dftb2.getEnergy(**scf_options)
        XYZ.write_xyz(opt_file, [atomlist_squeezed], title="Energy=%s" % en, mode="a")
        print "u = %20.10f  v = %20.10f  Energy = %20.10f" % (u,v,en)
        return en

    # optimize
    uv0 = np.array([1.05, 0.016])
    uv_opt = optimize.fmin(energy, uv0)
    print "Optimum:"
    print "u = %s  v = %s" % tuple(uv_opt)
    
    
